function [mass,F,ab,K,Stress,H]=evclus(D,c,lambda,version);
% EVCLUS Relational clustering - Metric approach
% 
% [mass,ab,K,Stress,H,F]=EVCLUS(D,c,lambda,version)
% Input
% 	D: dissimilarity matrix, size (n,n)
%   c: number of clusters
%	lambda: regularization parameter
%	version: choice of focal elements
%            0: only singletons
%            1: singletons + empty set + Omega (default)
%            >1 : all the 2^c focal elements
%
% Output
%	mass: mass assignment to each focal element, size (n,f), f=nb of focal elements
%   F: matrix (f,c) of focal elements
%      F(i,j)=1 if omega_j belongs to focal element i
%             0 otherwise
%	ab: coefficients in the Stress function
%        ab(1)= slope (=a)
%        ab(2)=constant (=b)
%	K: matrix (n,n) of degrees of conflict
%      K(i,j)=degree of conflict between m_i and m_j
%	Stress : stress function (without regularization term)
%	H : entropy (regularization term)
%
% References:
% M. Masson and T. Denoeux. Clustering of proximity data using belief functions.
%   IPMU'2002, Annecy, France, July 2002.
% T. Denoeux and M. Masson. EVCLUS: Evidential Clustering of Proximity Data.
%   Submitted to IEEE Trans. SMC B, 2002.

%
%   Version : Apr 12,03
%   Copyright (c) 2003 by T. Denoeux and M. Masson
% 	Author: T. Denoeux and M. Masson
%

[n1,n2]=size(D);
if n1~=n2
    error('Matrix D must be a squared matrix');
else,
    n=n1;
end;

if nargin ==3,
   version=1; % by default, the focals elements are the singletons+ emptyset
   %            + Omega
elseif nargin<3,
   disp('ERROR: evclus requires at least 3 parameters')
end;

W=1-eye(n);
I=find(W);
D1=D(I); % the distances are arranged in a vector

if version==0,
   F = eye(c); % focal elements=singletons
elseif version==1,
   F=[eye(c);zeros(1,c);ones(1,c)];
   % focal elements = singletons + empty set + Omega
else, % general case, all the 2^c subsets of Omega
    % not recommended unless c very small
   ii=1:2^c;
   F=zeros(length(ii),c);
   for i=1:c, F(:,i)=bitget(ii'-1,i); end;
end;

f=size(F,1);
xi=zeros(f,f);  % the matrix used to compute the degrees of conflict
for i=1:f,
   for j=1:f,
      xi(i,j)=1-max(min(F([i j],:)));
   end;
end;

card=sum(F,2)'; % the cardinality of the subsets 
i0=find(card==0);
if ~isempty(i0), card(i0)=c;end; % the 'cardinality' of the empty set is
% defined as c, this simplifies the calculation of H

% Initialization of credal partition
alpha=0.1*randn(n,f);
K=conflict(alpha,xi);
K1=K(I); % the degrees of conflict are arranged in a vector, as the distances
ab=[mean(D1)./mean(K1);0]; % initialization of the coefficients to appear in
% the Stress function

fprintf('EVCLUS - Metric method\n');

% stress minimization

param=[reshape(alpha,n*f,1);ab(:)];
options=[1 2000 1e-6 20]; % parameters passed to the optimization routine,
%                           see below
P=harris('coutmds',param,options,D,n,lambda,xi,card,W);

alpha=reshape(P(1:n*f),n,f);
[K,mass]=conflict(alpha,xi);
ab=P(n*f+1:end);

% stress and entropy computation
K1=K(I);
C=1/sum(D1); % the normalizing constant in the Stress function
if min(D1)==0, cst=0.1*mean(D1); else, cst=0; end;
% this is necessary to avoid division by 0 in the Sammon's Stress function
Z=log(repmat(card,n,1)./(max(mass,eps)))/log(2);
H=sum(sum(mass .* Z))/n;  % the total uncertainty entropy measure
Stress=C*sum(((ab(1)*K1+ab(2)-D1).^2)./(D1+cst));

disp(['Normalized stress: ' num2str(Stress)])
disp(['Entropy: ' num2str(H)])


% ---------------------------------------------------------------------------------
function [x,y] = harris(fun,x,options,varargin)
%
% Harris optimization method
%
% fun: function to be optimized (usually an M-file fun.m)
% The function 'FUN' should return a scalar function value [F,G]=FUN(x) and 
% the partial derivatives of the function dfun/dx, at the point x.
%
% options(1) = Display parameter : 1 displays some results.
% options(2) = maximum number of iterations
% options(3) = error
% options(4) = number of iterations between two displays

pas = 0.1 * ones(size(x));
a = 1.2;
b = 0.8;
c = 0.5;
ovf = 1e4 ;
unf = 1e-6;

err=[];
it = 0;
gain=1;

[yp,gp]=feval(fun,x,varargin{:});

xp=x;
x = xp - pas .* gp;

while (gain >= options(3)) & (it <= options(2))
  it=it+1;
  [y,g]=feval(fun,x,varargin{:});
  if options(1) >0,
    if rem(it,options(4))==1, 
      fprintf(1,'%5.0f %7.4f %6.4g \n',[it y gain]); 
    end;
  end;
  if y > yp,    
    x = xp;
    g = gp; 
    pas = pas * c;
    x = x - pas .* g;
  else
    gain=.9*gain+.1*abs(yp-y);
    xp = x;
    test = (g .* gp) >= 0;
    pas = ((test * a) + ((~test) * b)) .* pas;
    pas = (pas<=ovf) .* pas + (pas>ovf) * ovf;
    pas = (pas>=unf) .* pas + (pas<unf) * unf;
    gp = g;
    x = x - pas .* g;
    yp = y;
  end;
end

% ---------------------------------------------------------------------------------
function [J,gradJ]=coutmds(param,D,n,lambda,xi,card,W);
% 
% The function called by the optimization routine harris
% Computes the objective function J = Stress + lambda*H
% and its derivatives w.r.t alpha(i,k), a and b
% Input:
% param: alpha(i,k), i=1,..,n, k=1,..,f, a and b
% D: dissimialrities
% n: number of objects
% lambda: regularization parameter
% xi: matrix (f,f) used to compute the degrees of conflict
% card: cardinalities of the focal elements
% W : a matrix (n,n) with 0's on the diagonal, and 1's elsewhere
%
% Output:
% J: objective function 
% gradJ: vector (nf+2,1), gradient of J

f=size(xi,1);  % number of focal elements
alpha=reshape(param(1:n*f),n,f);
[K,mass]=conflict(alpha,xi);
ab=param(n*f+1:end);

% Calculation of objective function
I=find(W);
K1=K(I);
D1=D(I);
C=1/sum(D1); % the normalizing constant in the Stress function
if min(D1)==0, cst=0.1*mean(D1); else, cst=0; end;
% this is necessary to avoid division by 0 in the Sammon's Stress function
D=D+100*eye(n);

Z=log(repmat(card,n,1)./(max(mass,eps)))/log(2);
H=sum(sum(mass .* Z))/n; % the total uncertainty entropy measure
J=C*sum(((ab(1)*K1+ab(2)-D1).^2)./(D1+cst))+lambda*H;

% Calculation of the gradient
gradalpha=zeros(n,f);
for k=1:n,
	ek=W(k,:).*(ab(1)*K(k,:)+ab(2) - D(k,:));
    ek=ek./(D(k,:)+cst);
	ek(k)=0;
	ek=ek';
	A=mass(k,:)'*mass(k,:);
	v = mass(k,:).*(mass(k,:)-1);
	A=A.*(1-eye(f))+diag(v);	
	B=xi*mass';
	dsk=B'*A;
	dsk=ab(1)*dsk.*repmat(ek,1,f);
    G=sum(dsk,1);
	PartOne = 4*C*G; 
    dHdmk=Z(k,:)-1;
    gradH=dHdmk*A/n;
    gradalpha(k,:)=PartOne + lambda*gradH;	
end;
	
grada=2*C*sum((ab(1)*K1+ab(2)-D1).*K1./(D1+cst));
gradb=2*C*sum((ab(1)*K1+ab(2)-D1)./(D1+cst));
gradJ=[reshape(gradalpha,n*f,1);grada;gradb];

%---------------------------------------------------------------------
function [K,mass]=conflict(alpha,xi)

% Calculation of the degrees of conflict for a credal partition alpha
% Input
% alpha: parameters defiing the credal partitions, size (n,f)
% xi: matrix (f,f) for computing the degrees of conflict
%
% Output
% K: matrix (n,n), degrees of conflict
% mass: the bba's, matrix (n,f)

f=size(alpha,2);
mass=exp(-alpha);
mass=mass./repmat(sum(mass,2),1,f);
K=mass*xi*mass';

